from typing import Tuple
from copy import deepcopy

import numpy as np
import torch
from torch.utils.data import DataLoader
import torchvision.transforms as transforms
import torchvision as tv

from experiment_config import Config, DatasetType, HParams


# Return two batches, each containing #sample_size datapoints
# These form #sample_size pairs
def get_stiffness_batches(loader1, loader2, sample_size):
    # batch1 - batch1_prime : train - train
    # batch1 - batch2 : train - val
    loader1_iter = iter(DataLoader(loader1.dataset, batch_size=sample_size, shuffle=True, num_workers=0,
                                   collate_fn=None))
    batch1 = next(loader1_iter)
    batch1_prime = next(loader1_iter)
    batch2 = next(iter(DataLoader(loader2.dataset, batch_size=sample_size, shuffle=True, num_workers=0,
                                  collate_fn=None)))
    return batch1, batch1_prime, batch2


def default_collate(imgs):
    tensorise = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])
    all_tensors = []
    all_targets = []
    for img, target in imgs:
        tensorised_img = tensorise(img).unsqueeze(0)
        all_tensors.append(tensorised_img)
        all_targets.append(target)
    return torch.cat(all_tensors, dim=0), torch.LongTensor(all_targets)


def replicate_and_transform(hparams: HParams):
    assert hparams.K_train > 0

    normal_transformations = transforms.Compose([
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])

    def custom_collate(imgs):
        all_tensors = []
        all_targets = []
        for img, target in imgs:
            replicated_and_transformed = [normal_transformations(deepcopy(img)).unsqueeze(0) for _ in range(hparams.K_train)]
            img_tensor = torch.cat(replicated_and_transformed, dim=0)
            all_tensors.append(img_tensor)
            all_targets.append(target)
        return torch.cat(all_tensors, dim=0), torch.LongTensor(all_targets)

    return custom_collate


def get_dataloaders(hparams: HParams, config: Config, device: torch.device) -> Tuple[
    DataLoader, DataLoader, DataLoader]:
    # if hparams.dataset_type == DatasetType.CIFAR10:
    #     dataset = CIFAR10
    # elif hparams.dataset_type == DatasetType.SVHN:
    #     dataset = SVHN
    # else:
    #     raise KeyError

    # train_key = {'split': 'train'} if hparams.dataset_type == DatasetType.SVHN else {'train': True}
    # test_key = {'split': 'test'} if hparams.dataset_type == DatasetType.SVHN else {'train': False}
    # train = dataset(hparams, config, device, download=True, **train_key)
    # test = dataset(hparams, config, device, download=True, **test_key)

    if hparams.K_train > 0:
        transform_train = transforms.Compose([])
        transform_test = transforms.Compose([])
        collate_function = replicate_and_transform(hparams)
    else:
        if hparams.transforms:
            print("Using transformations")
            transform_train = transforms.Compose([
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])
        else:
            print("No transformations")
            transform_train = transforms.Compose([
                # transforms.RandomCrop(32, padding=4),
                # transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])
        transform_test = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ])
        collate_function = None

    trainset = tv.datasets.CIFAR10(
        root='./temp/data', train=True, download=True, transform=transform_train)
    testset = tv.datasets.CIFAR10(
        root='./temp/data', train=False, download=True, transform=transform_test)

    train_loader = DataLoader(
        trainset, batch_size=hparams.batch_size, shuffle=True, collate_fn=collate_function)
    test_loader = DataLoader(testset, batch_size=1250, shuffle=False, collate_fn=collate_function)
    evalset = tv.datasets.CIFAR10(root='./temp/data', train=True, download=True, transform=transform_test)
    train_eval_loader = DataLoader(evalset, batch_size=1250, shuffle=False, collate_fn=collate_function)
    return train_loader, train_eval_loader, test_loader


def process_data(hparams: HParams, data_np: np.ndarray, targets_np: np.ndarray, device: torch.device, train: bool):
    # Scale data to [0,1] floats
    data_np = data_np / 255

    # Normalize data
    data_np = (data_np - data_np.mean(axis=(0, 1, 2))) / data_np.std(axis=(0, 1, 2))

    # NHWC -> NCHW
    data_np = data_np.transpose((0, 3, 1, 2))

    # Numpy -> Torch
    data = torch.tensor(data_np, dtype=torch.float32)
    targets = torch.tensor(targets_np, dtype=torch.long)

    # Resize dataset
    dataset_size, offset = (hparams.train_dataset_size, 0) if train else (hparams.test_dataset_size, 1)
    if dataset_size is not None:
        rng = np.random.RandomState(hparams.data_seed + offset) if (hparams.data_seed is not None) else np.random
        indices = rng.choice(len(data), dataset_size, replace=False)
        indices = torch.from_numpy(indices)
        data = torch.index_select(data, 0, indices)
        targets = torch.index_select(targets, 0, indices)

    # Put both data and targets on GPU in advance
    return data.to(device), targets.to(device)


# https://gist.github.com/y0ast/f69966e308e549f013a92dc66debeeb4
# We need to keep the class name the same as base class methods rely on it
class CIFAR10(tv.datasets.CIFAR10):
    def __init__(self, hparams: HParams, config: Config, device: torch.device, *args, **kwargs):
        super().__init__(config.data_dir, *args, **kwargs)
        self.data, self.targets = process_data(hparams, self.data, np.array(self.targets), device, self.train)

    # Don't convert to PIL like torchvision default
    def __getitem__(self, index):
        return self.data[index], self.targets[index]


class SVHN(tv.datasets.SVHN):
    def __init__(self, hparams: HParams, config: Config, device: torch.device, *args, **kwargs):
        super().__init__(config.data_dir, *args, **kwargs)
        self.data = self.data.transpose((0, 2, 3, 1))  # NCHW -> NHWC (SVHN)
        self.data, self.labels = process_data(hparams, self.data, self.labels, device, self.split == 'train')

    def __getitem__(self, index):
        return self.data[index], self.labels[index]
